"""
Utility functions to mimic the template support functions for vtkVariant
"""

from vtkmodules import vtkCommonCore
import sys

_variant_type_map = {
    'void' : vtkCommonCore.VTK_VOID,
    'char' : vtkCommonCore.VTK_CHAR,
    'unsigned char' : vtkCommonCore.VTK_UNSIGNED_CHAR,
    'signed char' : vtkCommonCore.VTK_SIGNED_CHAR,
    'short' : vtkCommonCore.VTK_SHORT,
    'unsigned short' : vtkCommonCore.VTK_UNSIGNED_SHORT,
    'int' : vtkCommonCore.VTK_INT,
    'unsigned int' : vtkCommonCore.VTK_UNSIGNED_INT,
    'long' : vtkCommonCore.VTK_LONG,
    'unsigned long' : vtkCommonCore.VTK_UNSIGNED_LONG,
    'long long' : vtkCommonCore.VTK_LONG_LONG,
    'unsigned long long' : vtkCommonCore.VTK_UNSIGNED_LONG_LONG,
    'float' : vtkCommonCore.VTK_FLOAT,
    'double' : vtkCommonCore.VTK_DOUBLE,
    'string' : vtkCommonCore.VTK_STRING,
    'vtkObjectBase' : vtkCommonCore.VTK_OBJECT,
    'vtkObject' : vtkCommonCore.VTK_OBJECT,
}

_variant_method_map = {
    vtkCommonCore.VTK_VOID : '',
    vtkCommonCore.VTK_CHAR : 'ToChar',
    vtkCommonCore.VTK_UNSIGNED_CHAR : 'ToUnsignedChar',
    vtkCommonCore.VTK_SIGNED_CHAR : 'ToSignedChar',
    vtkCommonCore.VTK_SHORT : 'ToShort',
    vtkCommonCore.VTK_UNSIGNED_SHORT : 'ToUnsignedShort',
    vtkCommonCore.VTK_INT : 'ToInt',
    vtkCommonCore.VTK_UNSIGNED_INT : 'ToUnsignedInt',
    vtkCommonCore.VTK_LONG : 'ToLong',
    vtkCommonCore.VTK_UNSIGNED_LONG : 'ToUnsignedLong',
    vtkCommonCore.VTK_LONG_LONG : 'ToLongLong',
    vtkCommonCore.VTK_UNSIGNED_LONG_LONG : 'ToUnsignedLongLong',
    vtkCommonCore.VTK_FLOAT : 'ToFloat',
    vtkCommonCore.VTK_DOUBLE : 'ToDouble',
    vtkCommonCore.VTK_STRING : 'ToString',
    vtkCommonCore.VTK_OBJECT : 'ToVTKObject',
}

_variant_check_map = {
    vtkCommonCore.VTK_VOID : 'IsValid',
    vtkCommonCore.VTK_CHAR : 'IsChar',
    vtkCommonCore.VTK_UNSIGNED_CHAR : 'IsUnsignedChar',
    vtkCommonCore.VTK_SIGNED_CHAR : 'IsSignedChar',
    vtkCommonCore.VTK_SHORT : 'IsShort',
    vtkCommonCore.VTK_UNSIGNED_SHORT : 'IsUnsignedShort',
    vtkCommonCore.VTK_INT : 'IsInt',
    vtkCommonCore.VTK_UNSIGNED_INT : 'IsUnsignedInt',
    vtkCommonCore.VTK_LONG : 'IsLong',
    vtkCommonCore.VTK_UNSIGNED_LONG : 'IsUnsignedLong',
    vtkCommonCore.VTK_LONG_LONG : 'IsLongLong',
    vtkCommonCore.VTK_UNSIGNED_LONG_LONG : 'IsUnsignedLongLong',
    vtkCommonCore.VTK_FLOAT : 'IsFloat',
    vtkCommonCore.VTK_DOUBLE : 'IsDouble',
    vtkCommonCore.VTK_STRING : 'IsString',
    vtkCommonCore.VTK_OBJECT : 'IsVTKObject',
}


def vtkVariantCreate(v, t):
    """
    Create a vtkVariant of the specified type, where the type is in the
    following format: 'int', 'unsigned int', etc. for numeric types,
    and 'string' for strings.  You can also use an
    integer VTK type constant for the type.
    """
    if not issubclass(type(t), int):
        t = _variant_type_map[t]

    return vtkCommonCore.vtkVariant(v, t)


def vtkVariantExtract(v, t=None):
    """
    Extract the specified value type from the vtkVariant, where the type is
    in the following format: 'int', 'unsigned int', etc. for numeric types,
    and 'string' for strings.  You can also use an
    integer VTK type constant for the type.  Set the type to 'None" to
    extract the value in its native type.
    """
    v = vtkCommonCore.vtkVariant(v)

    if t == None:
        t = v.GetType()
    elif not issubclass(type(t), int):
        t = _variant_type_map[t]

    if getattr(v, _variant_check_map[t])():
        return getattr(v, _variant_method_map[t])()
    else:
        return None


def vtkVariantCast(v, t):
    """
    Cast the vtkVariant to the specified value type, where the type is
    in the following format: 'int', 'unsigned int', etc. for numeric types,
    and 'string' for strings.  You can also use an
    integer VTK type constant for the type.
    """
    if not issubclass(type(t), int):
        t = _variant_type_map[t]

    v = vtkCommonCore.vtkVariant(v, t)

    if v.IsValid():
        return getattr(v, _variant_method_map[t])()
    else:
        return None


def vtkVariantStrictWeakOrder(s1, s2):
    """
    Compare variants by type first, and then by value.
    """
    s1 = vtkCommonCore.vtkVariant(s1)
    s2 = vtkCommonCore.vtkVariant(s2)

    t1 = s1.GetType()
    t2 = s2.GetType()

    # check based on type
    if t1 != t2:
        return t1 < t2

    v1 = s1.IsValid()
    v2 = s2.IsValid()

    # check based on validity
    if (not v1) or (not v2):
        return v1 < v2

    # extract and compare the values
    r1 = getattr(s1, _variant_method_map[t1])()
    r2 = getattr(s2, _variant_method_map[t2])()

    # compare vtk objects by classname, then address
    if t1 == vtkCommonCore.VTK_OBJECT:
        c1 = r1.GetClassName()
        c2 = r2.GetClassName()
        if c1 != c2:
            return c1 < c2
        else:
            return r1.__this__ < r2.__this__

    return r1 < r2


class vtkVariantStrictWeakOrderKey:
    """A key method (class, actually) for use with sort()"""
    def __init__(self, obj, *args):
        self.obj = obj
    def __lt__(self, other):
        return vtkVariantStrictWeakOrder(self.obj, other)


def vtkVariantStrictEquality(s1, s2):
    """
    Check two variants for strict equality of type and value.
    """
    s1 = vtkCommonCore.vtkVariant(s1)
    s2 = vtkCommonCore.vtkVariant(s2)

    t1 = s1.GetType()
    t2 = s2.GetType()

    # check based on type
    if t1 != t2:
        return False

    v1 = s1.IsValid()
    v2 = s2.IsValid()

    # check based on validity
    if (not v1) and (not v2):
        return True
    elif v1 != v2:
        return False

    # extract and compare the values
    r1 = getattr(s1, _variant_method_map[t1])()
    r2 = getattr(s2, _variant_method_map[t2])()

    return (r1 == r2)


def vtkVariantLessThan(s1, s2):
    """
    Return true if s1 < s2.
    """
    return (vtkCommonCore.vtkVariant(s1) < vtkCommonCore.vtkVariant(s2))


def vtkVariantEqual(s1, s2):
    """
    Return true if s1 == s2.
    """
    return (vtkCommonCore.vtkVariant(s1) == vtkCommonCore.vtkVariant(s2))



import cftime
import logging
import numpy as np
from os.path import basename, splitext, exists
import xarray as xr
from vtkmodules.vtkCommonCore import (
    vtkVariant,
)
from vtkmodules.vtkCommonDataModel import (
    vtkDataObject
)
from vtkmodules.vtkCommonExecutionModel import (
    vtkAlgorithm,
    vtkStreamingDemandDrivenPipeline
)
from vtkmodules.vtkIONetCDF import vtkNetCDFCFReader, vtkXArrayAccessor
from vtkmodules.util import numpy_support
from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase

@xr.register_dataset_accessor("vtk")
class VtkAccessor:
    def __init__(self, dsxr):
        self._dsxr = dsxr

    def create_reader(self):
        '''
        Returns a vtkXArrayCFReader that reads data from the XArray
        (using zero-copy when possible). At the moment, data is copied
        for coordinates (because they are converted to double in the reader)
        and for certain data that is subset either in XArray or in VTK.
        Lazy loading in XArray is respected, that is data is accessed only when
        it is needed.
        Time is passed to VTK either as an int64 for datetime64 or timedelta64,
        or as a double (using cftime.toordinal) for cftime.
        '''
        reader = vtkXArrayCFReader()
        reader.SetXArray(self._dsxr)
        return reader


class vtkXArrayCFReader(VTKPythonAlgorithmBase):
    '''Reads data from a file using the XArray readers and then connects
    the XArray data to the vtkNetCDFCFREader (using zero-copy when
    possible). At the moment, data is copied for coordinates (because
    they are converted to double in the reader) and for certain data
    that is subset either in XArray or in VTK.  Lazy loading in XArray
    is respected, that is data is accessed only when it is needed.
    Time is passed to VTK either as an int64 for datetime64 or
    timedelta64, or as a double (using cftime.toordinal) for cftime.
    '''

    _FORWARD_GET = {
        "GetAccessor",
        "GetAllDimensions",

        "GetNumberOfVariableArrays",
        "GetAllVariableArrayNames",
        "GetVariableArrayName",
        "GetVariableArrayStatus",

        "GetTimeDimensionName",
        "GetLatitudeDimensionName",
        "GetLongitudeDimensionName",
        "GetVerticalDimensionName",

        "GetOutput",
        "GetOutputType",
        "GetSphericalCoordinates",

        "GetReplaceFillValueWithNan",

        "GetVariableDimensions",
        "GetVerticalBias",
        "GetVerticalScale",
        "PrintSelf",
    }
    _FORWARD_SET = {
        "SetDimensions",

        "SetTimeDimensionName",
        "SetLatitudeDimensionName",
        "SetLongitudeDimensionName",
        "SetVerticalDimensionName",


        "SetSphericalCoordinates",
        "SphericalCoordinatesOn",
        "SphericalCoordinatesOff",

        "SetReplaceFillValueWithNan",
        "ReplaceFillValueWithNanOn",
        "ReplaceFillValueWithNanOff",

        "SetOutputType",
        "SetOutputTypeToAutomatic",
        "SetOutputTypeToImage",
        "SetOutputTypeToRectilinear",
        "SetOutputTypeToStructured",
        "SetOutputTypeToUnstructured",

        "SetVariableArrayStatus",
        "SetVerticalBias",
        "SetVerticalScale",
        "UpdateMetaData",
    }

    def __init__(self):
        VTKPythonAlgorithmBase.__init__(
            self, nInputPorts=0, nOutputPorts=1, outputType="vtkDataObject"
        )
        self._log = logging.getLogger("vtkXArrayCFReader")
        self._filename = None
        self._timesteps = None
        self._timeindex = None
        self._node = None
        self._dsxr = None
        self._reader = vtkNetCDFCFReader()
        self._ndarray_cftime_toordinal = np.frompyfunc(vtkXArrayCFReader._cftime_toordinal, 1, 1)
        # reference to contiguous arrays so that they are not dealocated
        self._arrays = {}


    def __getattr__(self, name):
        in_set = name in self._FORWARD_SET
        in_get = name in self._FORWARD_GET
        if in_set or in_get:
            if in_set:
                self.Modified()
            return getattr(self._reader, name)
        else:
            raise AttributeError()

    def SetFileName(self, name):
        """Specify filename for the file to read."""
        if self._filename != name:
            self._filename = name
            self.Modified()

    def GetFileName(self):
        return self._filename

    def CanReadFile(self, filepath):
        ext = splitext(filepath)[1]
        filename = basename(filepath)
        correct_name = False
        if ext == '.nc' or ext == '.grib' or ext == '.h5':
            correct_name = True
        else:
            if ext == '' and filename == '.zgroup':
                correct_name = True
        if correct_name and exists(filepath):
            return 1
        else:
            return 0


    def SetNode(self, node):
        if self._node != node:
            self._node = node
            self.Modified()

    def GetNode(self):
        return self._node

    def SetXArray(self, dsxr):
        self._dsxr = dsxr
        self._update_accessor()
        self.Modified()

    def GetXArray(self):
        return self._dsxr

    def RequestDataObject(self, request, inInfo, outInfo):
        self._log.debug(f"DataObject ======================================================================")
        if not self._dsxr:
            if self._node:
                tree = xr.open_datatree(self._filename)
                self._dsxr = tree[self._node].to_dataset()
            else:
                self._dsxr = xr.open_dataset(self._filename, decode_timedelta=True)
            self._update_accessor()
        self._reader.UpdateDataObject()
        roi = self._reader.GetOutputInformation(0)
        if roi.Has(vtkDataObject.DATA_OBJECT()):
            rdata = roi.Get(vtkDataObject.DATA_OBJECT())
        else:
            self._log.error("vtkNetCDFCFReader did not create the dataset")
            rdata = None
        oi = outInfo.GetInformationObject(0)
        oi.Set(vtkDataObject.DATA_OBJECT(), rdata)
        return 1

    def RequestInformation(self, request, inInfo, outInfo):
        self._log.debug(f"Information ======================================================================")
        oi = outInfo.GetInformationObject(0)
        self._reader.UpdateInformation()
        roi = self._reader.GetOutputInformation(0)
        if roi.Has(vtkStreamingDemandDrivenPipeline.TIME_STEPS()):
            self._timesteps = roi.Get(vtkStreamingDemandDrivenPipeline.TIME_STEPS())
            oi.Set(vtkStreamingDemandDrivenPipeline.TIME_STEPS(), self._timesteps, len(self._timesteps))
            oi.Set(vtkStreamingDemandDrivenPipeline.TIME_RANGE(), [self._timesteps[0], self._timesteps[-1]], 2)
            self._timesteps = np.asarray(self._timesteps)
        if roi.Has(vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT()):
            ext = roi.Get(vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT())
            self._log.debug("Whole extent: {}".format(ext))
            oi.Set(vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT(), ext, 6)
        if roi.Has(vtkAlgorithm.CAN_HANDLE_PIECE_REQUEST()):
            oi.Set(vtkAlgorithm.CAN_HANDLE_PIECE_REQUEST(), 1)
        if roi.Has(vtkAlgorithm.CAN_PRODUCE_SUB_EXTENT()):
            oi.Set(vtkAlgorithm.CAN_PRODUCE_SUB_EXTENT(), 1)
        return 1

    def RequestUpdateExtent(self, request, inInfo, outInfo):
        self._log.debug(f"UpdateExtent ======================================================================")
        oi = outInfo.GetInformationObject(0)
        if oi.Has(vtkStreamingDemandDrivenPipeline.UPDATE_TIME_STEP()):
            utime = oi.Get(vtkStreamingDemandDrivenPipeline.UPDATE_TIME_STEP())
            timeindex = (np.abs(self._timesteps - utime)).argmin()
            if timeindex != self._timeindex:
                self._log.debug(f"Time index = {timeindex}")
                self._timeindex = timeindex
                self.Modified()
        if oi.Has(vtkStreamingDemandDrivenPipeline.UPDATE_EXTENT()):
            ext = [0, 0, 0, 0, 0, 0]
            oi.Get(vtkStreamingDemandDrivenPipeline.UPDATE_EXTENT(), ext)
            self._log.debug("Update extent: {}".format(ext))
            roi = self._reader.GetOutputInformation(0)
            roi.Set(vtkStreamingDemandDrivenPipeline.UPDATE_EXTENT(), ext, 6)
            self._reader.PropagateUpdateExtent()
        return 1


    def RequestData(self, request, inInfo, outInfo):
        self._log.debug(f"Data ======================================================================")
        if self._timeindex:
            dsxr = self._dsxr.isel({self.GetTimeDimensionName() : self._timeindex})
        else:
            # no time, so no aditional selection is needed
            dsxr = self._dsxr
        accessor = self._reader.GetAccessor()
        self._set_data_vars(accessor, dsxr)
        self._reader.Update()
        # self._reader's data is already set for this's data so no ShallowCopy is needed
        return 1

    @staticmethod
    def _get_nc_type(numpy_array_type):
        """Returns a nc_type given a numpy array."""
        NC_BYTE = 1  # 1 byte integer
        NC_CHAR = 2  # iso/ascii character
        NC_SHORT = 3  # 2 byte integer
        NC_INT = 4  # 4 byte integer
        NC_LONG = NC_INT
        NC_FLOAT = 5
        NC_DOUBLE = 6
        NC_UBYTE = 7
        NC_USHORT = 8
        NC_UINT = 9
        NC_INT64 = 10  # 8 bypte integer
        NC_UINT64 = 11
        NC_STRING = 12
        _np_nc = {
            np.uint8: NC_UBYTE,
            np.uint16: NC_USHORT,
            np.uint32: NC_UINT,
            np.uint64: NC_UINT64,
            np.int8: NC_BYTE,
            np.int16: NC_SHORT,
            np.int32: NC_INT,
            np.int64: NC_INT64,
            np.float32: NC_FLOAT,
            np.float64: NC_DOUBLE,
            np.datetime64: NC_INT64,
            np.timedelta64: NC_INT64,
            np.str_: NC_STRING,
            np.bytes_: NC_CHAR,
        }
        for key, nc_type in _np_nc.items():
            if (
                numpy_array_type == key
                or np.issubdtype(numpy_array_type, key)
                or numpy_array_type == np.dtype(key)
            ):
                return nc_type
        raise TypeError(
            "Could not find a suitable NetCDF type for %s" % (str(numpy_array_type))
        )

    def _update_accessor(self):
        accessor, timename = self._get_accessor()
        self._reader.SetAccessor(accessor)
        if timename:
            self._reader.SetTimeDimensionName(timename)

    def _get_accessor(self):
        acclog = logging.getLogger("_get_accessor_")
        acclog.setLevel(logging.WARNING)
        accessor = vtkXArrayAccessor()
        time_name = None
        time_names = []
        # Set Dim and DimLen
        dimNameToIndex = {k: i for i, k in enumerate(self._dsxr.sizes.keys())}
        accessor.SetDim(list(self._dsxr.sizes.keys()))
        accessor.SetDimLen(list(self._dsxr.sizes.values()))

        # Set Var
        varList = list(self._dsxr.data_vars.keys()) + list(self._dsxr.coords.keys())
        varNameToIndex = {k: i for i, k in enumerate(varList)}
        is_coord = [0] * len(self._dsxr.data_vars)
        is_coord = is_coord + [1] * len(self._dsxr.coords)
        coords_bounds = self._get_coords_bounds()
        accessor.SetVar(varList, is_coord)
        for i, v in enumerate(varList):
            # data_vars are set after array selection and time selection to
            # take advantage of xarray lazy loading
            # https://docs.xarray.dev/en/latest/internals/internal-design.html
            if is_coord[i] or v in coords_bounds:
                # if there is subsetting in xarray, self._dsxr[v].values is
                # not contiguous. If the array is not contiguous, a contiguous
                # copy is created otherwise the contiguous array is simply returned
                v_data = np.ascontiguousarray(self._dsxr[v].values)
                if (
                    v_data.dtype.type == np.datetime64
                    or v_data.dtype.type == np.timedelta64
                ):
                    un = np.datetime_data(v_data.dtype)
                    # unit = ns and 1 base unit
                    if un[0] == "ns" and un[1] == 1:
                        time_names.append(v)
                if v_data.dtype.char == "O":
                    # object array, assume cftime
                    # copy cftime array to a doubles array
                    self._arrays[v] = self._ndarray_cftime_toordinal(v_data).astype(np.float64)
                    time_names.append(v)
                    v_data = self._arrays[v]
                else:
                    self._arrays[v] = v_data
                acclog.debug(f"{v=} {v_data.shape=} {v_data.dtype} {self._dsxr[v].dims=}")
                acclog.debug(f"address:{hex(v_data.ctypes.data)}")
                accessor.SetVarValue(i, v_data)
                accessor.SetVarType(i, vtkXArrayCFReader._get_nc_type(v_data.dtype))
            else:
                accessor.SetVarType(i, vtkXArrayCFReader._get_nc_type(self._dsxr[v].variable.dtype))
            accessor.SetVarDims(i, [dimNameToIndex[name] for name in self._dsxr[v].dims])
            accessor.SetVarCoords(
                i, [varNameToIndex[name] for name in self._dsxr[v].coords]
            )

            acclog.debug("Attributes:")
            for item in self._dsxr[v].attrs.items():
                acclog.debug(
                    "name: {} value: {} type: {}".format(
                        item[0], item[1], type(item[1])
                    )
                )
                if np.issubdtype(type(item[1]), np.integer):
                    accessor.SetAtt(i, item[0], vtkVariant(int(item[1])))
                elif np.issubdtype(type(item[1]), np.floating):
                    accessor.SetAtt(i, item[0], vtkVariant(float(item[1])))
                elif isinstance(item[1], np.ndarray):
                    accessor.SetAtt(
                        i, item[0], vtkVariant(numpy_support.numpy_to_vtk(item[1]))
                    )
                else:
                    accessor.SetAtt(i, item[0], vtkVariant(item[1]))
        if len(time_names) >= 1:
            for name in time_names:
                if accessor.IsCOARDSCoordinate(name):
                    time_name = name
                    break
        return accessor, time_name

    def _set_data_vars(self, accessor, dsxr):
        # data_vars are listed first in the list of data_vars,coords so we don't
        # need to add coords to the list, and still get the corect indexes
        varList = list(dsxr.data_vars.keys())
        for i, v in enumerate(varList):
            if self._reader.GetVariableArrayStatus(v):
                v_data = np.ascontiguousarray(dsxr[v].values)
                accessor.SetVarValue(i, v_data)
                self._arrays[v] = v_data

    def _get_coords_bounds(self):
        '''
        Special data_vars associated coords
        '''
        b=set()
        for coord in list(self._dsxr.coords):
            bounds_attr = 'bounds'
            if bounds_attr in self._dsxr[coord].attrs:
                b.add(self._dsxr[coord].attrs[bounds_attr])
        return b


    @staticmethod
    def _cftime_toordinal(o):
        return o.toordinal(fractional=True)



#!/usr/bin/env python
# This is the script that runs Python regression test scripts.
# The test script to be run must be the first argument.

from vtkmodules.vtkCommonCore import vtkMath
from vtkmodules.vtkTestingRendering import vtkTesting
import os
import sys
import re
import traceback

def _GetController():
    try:
        from vtkmodules.vtkParallelMPI import vtkMPIController
        controller = vtkMPIController();

        # If MPI was not initialized, we do not want to use MPI
        if not controller.GetCommunicator():
            return None
        return controller
    except:
        pass
    return None

def main(test_script):
    """Run a regression test, and compare the contents of the window against
    against a valid image.  This will use arguments from sys.argv to set the
    testing options via the vtkTesting class, run the test script, and then
    call vtkTesting.RegressionTest() to validate the image.  The return
    value will the one provided by vtkTesting.RegressionTest().
    """

    # find the first argument that's an option
    opt1 = 1
    while opt1 < len(sys.argv) and not sys.argv[opt1].startswith('-'):
        opt1 += 1

    # get the "-A" option, which isn't handled by vtkTester
    for i in range(opt1, len(sys.argv)):
        if sys.argv[i] == '-A' and i < len(sys.argv) - 1:
            sys.path.append(sys.argv[i + 1])

    # create the testing class to do the work
    rtTester = vtkTesting()

    try:
        rtTester.SetController(_GetController())
    except:
        pass

    for arg in sys.argv[opt1:]:
        rtTester.AddArgument(arg)

    # if test is not interactive, make a mock interactor with a
    # disabled Start method
    if rtTester.IsInteractiveModeSpecified() == 0:
        from vtkmodules.vtkRenderingCore import vtkRenderWindowInteractor
        import vtkmodules.vtkRenderingUI
        irenType = type(vtkRenderWindowInteractor())
        class vtkTestingInteractor(irenType):
            def Start(self):
                pass
        irenType.override(vtkTestingInteractor)

    # seed the random number generator
    vtkMath.RandomSeed(6)

    # read the test script
    with open(test_script) as test_file:
        test_code = test_file.read()

    # inject the test script's directory into sys.path
    test_script_dir = os.path.abspath(os.path.dirname(test_script))
    sys.path.insert(0, test_script_dir)

    # we provide an initial set of variables for the test script
    test_script_vars = { "__name__" : "__main__" }

    try:
        # run the test and capture all of its global variables
        exec(compile(test_code, test_script, 'exec'), test_script_vars)
    except Exception:
        traceback.print_exc()
        return vtkTesting.FAILED
    finally:
        # undo the change to the path
        sys.path.remove(test_script_dir)

    # undo the vtkRenderWindowInteractor override
    if rtTester.IsInteractiveModeSpecified() == 0:
       irenType.override(None)

    # if script has "if __name__ == '__main__':", then we assume that
    # it would have raised an exception or called exit(1) if it failed.
    if re.search('^if *__name__ *== *[\'\"]__main__[\'\"]', test_code, flags=re.MULTILINE):
        return vtkTesting.PASSED

    # if script didn't set "threshold", use default value
    try:
        threshold = test_script_vars["threshold"]
    except KeyError:
        # Used to be 0.15. Changed to 0.05 for new SSIM method.
        threshold = 0.05

    # we require a valid regression image
    if rtTester.IsValidImageSpecified():
        # look for a renderWindow ImageWindow or ImageViewer
        # first check for some common names
        if "iren" in test_script_vars:
            iren = test_script_vars["iren"]
            rtTester.SetRenderWindow(iren.GetRenderWindow())
            iren.GetRenderWindow().Render()
        elif "renWin" in test_script_vars:
            renWin = test_script_vars["renWin"]
            rtTester.SetRenderWindow(renWin)
        elif "viewer" in test_script_vars:
            viewer = test_script_vars["viewer"]
            rtTester.SetRenderWindow(viewer.GetRenderWindow())
            viewer.Render()
        elif "imgWin" in test_script_vars:
            imgWin = test_script_vars["imgWin"]
            rtTester.SetRenderWindow(imgWin)
            imgWin.Render()
        else:
            sys.stderr.write("Test driver rtImageTest.py says \"no iren, renWin, viewer, or imgWin\": %s\n" % test_script)
            return vtkTesting.FAILED

        return rtTester.RegressionTest(threshold)

    return vtkTesting.FAILED

if __name__ == '__main__':
    # We don't parse the arguments (vtkTesting does that), but we need
    # to extract the name of the test script to run.
    if len(sys.argv) < 2 or sys.argv[1].startswith('-'):
        print("Usage %s <test script> [<addition arguments>]" % argv[0])
        sys.exit(1)

    if main(sys.argv[1]) == vtkTesting.FAILED:
        sys.exit(1)



"""
This internal module provides a number of algorithms designed to be
compatible with and used by numpy. In particular, some algorithms
required to be adapted to handle VTKCompositeDataArray. Note that
the decorated functions (by _override_numpy) are not meant to be
called directly from here but directly through the Numpy API.
"""
import numpy
from typing import Optional

from . import dataset_adapter as dsa
from . import internal_algorithms as algs

def _apply_func2(func, array, args):
    """Apply a function to each member of a VTKCompositeDataArray.
    Returns a list of arrays.

    Note that this function is mainly for internal use by this module."""
    if array is dsa.NoneArray:
        return []
    res = []
    for a in array.Arrays:
        if a is dsa.NoneArray:
            res.append(dsa.NoneArray)
        else:
            res.append(func(a, *args))
    return res

def apply_ufunc(func, array, args=()):
    """Apply a function to each member of a VTKCompositeDataArray.
    VTKArray and numpy arrays are also supported."""
    if array is dsa.NoneArray:
        return dsa.NoneArray
    elif type(array) == dsa.VTKCompositeDataArray:
        return dsa.VTKCompositeDataArray(_apply_func2(func, array, args), dataset = array.DataSet)
    else:
        return func(array, *args)

def _make_ufunc(ufunc):
    """ Given a ufunc, creates a closure that applies it to each member
    of a VTKCompositeDataArray.

    Note that this function is mainly for internal use by this module."""
    def new_ufunc(array, *args):
        return apply_ufunc(ufunc, array, args)
    return new_ufunc

def apply_dfunc(dfunc, array1, val2):
    """Apply a two argument function to each member of a VTKCompositeDataArray
    and another argument The second argument can be a VTKCompositeDataArray, in
    which case a one-to-one match between arrays is assumed. Otherwise, the
    function is applied to the composite array with the second argument repeated.
    VTKArray and numpy arrays are also supported."""
    if type(array1) == dsa.VTKCompositeDataArray and type(val2) == dsa.VTKCompositeDataArray:
        res = []
        for a1, a2 in zip(array1.Arrays, val2.Arrays):
            if a1 is dsa.NoneArray or a2 is dsa.NoneArray:
                res.append(dsa.NoneArray)
            else:
                l = dsa.reshape_append_ones(a1, a2)
                res.append(dfunc(l[0], l[1]))
        return dsa.VTKCompositeDataArray(res, dataset = array1.DataSet)
    elif type(array1) == dsa.VTKCompositeDataArray:
        res = []
        for a in array1.Arrays :
            if a is dsa.NoneArray:
                res.append(dsa.NoneArray)
            else:
                l = dsa.reshape_append_ones(a, val2)
                res.append(dfunc(l[0], l[1]))
        return dsa.VTKCompositeDataArray(res, dataset = array1.DataSet)
    elif array1 is dsa.NoneArray:
        return dsa.NoneArray
    elif isinstance(val2, numpy.ndarray):
        return dfunc(array1, val2)
    else:
        l = dsa.reshape_append_ones(array1, val2)
        return dfunc(l[0], l[1])

def _make_dfunc(dfunc):
    """ Given a function that requires two arguments, creates a closure that
    applies it to each member of a VTKCompositeDataArray.

    Note that this function is mainly for internal use by this module."""
    def new_dfunc(array1, val2):
        return apply_dfunc(dfunc, array1, val2)
    return new_dfunc

def _make_dsfunc(dsfunc):
    """ Given a function that requires two arguments (one array, one dataset),
    creates a closure that applies it to each member of a VTKCompositeDataArray.
    Note that this function is mainly for internal use by this module."""
    def new_dsfunc(array, ds=None):
        if type(array) == dsa.VTKCompositeDataArray:
            res = []
            for a in array.Arrays:
                if a is dsa.NoneArray:
                    res.append(dsa.NoneArray)
                else:
                    res.append(dsfunc(a, ds))
            return dsa.VTKCompositeDataArray(res, dataset = array.DataSet)
        elif array is dsa.NoneArray:
            return dsa.NoneArray
        else:
            return dsfunc(array, ds)
    return new_dsfunc

def _make_dsfunc2(dsfunc):
    """ Given a function that requires a dataset, creates a closure that
    applies it to each member of a VTKCompositeDataArray.

    Note that this function is mainly for internal use by this module."""
    def new_dsfunc2(ds):
        if type(ds) == dsa.CompositeDataSet:
            res = []
            for dataset in ds:
                res.append(dsfunc(dataset))
            return dsa.VTKCompositeDataArray(res, dataset = ds)
        else:
            return dsfunc(ds)
    return new_dsfunc2

@dsa._override_numpy(numpy.bitwise_or)
def bitwise_or(array1, array2):
    """Implements element by element or (bitwise, | in C/C++) operation.
    If one of the arrays is a NoneArray, this will return the array
    that is not NoneArray, treating NoneArray as 0 in the or operation."""
    if type(array1) == dsa.VTKCompositeDataArray and type(array2) == dsa.VTKCompositeDataArray:
        res = []
        for a1, a2 in zip(array1.Arrays, array2.Arrays):
            l = dsa.reshape_append_ones(a1, a2)
            res.append(bitwise_or(l[0], l[1]))
        return dsa.VTKCompositeDataArray(res, dataset = array1.DataSet)
    elif type(array1) == dsa.VTKCompositeDataArray:
        res = []
        for a in array1.Arrays :
            l = dsa.reshape_append_ones(a, array2)
            res.append(bitwise_or(l[0], l[1]))
        return dsa.VTKCompositeDataArray(res, dataset = array1.DataSet)
    elif array1 is dsa.NoneArray:
        return array2
    elif array2 is dsa.NoneArray:
        return array1
    else:
        l = dsa.reshape_append_ones(array1, array2)
        return numpy.bitwise_or(l[0], l[1])

@dsa._override_numpy(numpy.sum)
def sum(array, axis=None, controller=None):
    """Returns the sum of all values along a particular axis (dimension).
    Given an array of m tuples and n components:
    * Default is to return the sum of all values in an array.
    * axis=0: Sum values of all components and return a one tuple,
      n-component array.
    * axis=1: Sum values of all components of each tuple and return an
      m-tuple, 1-component array.

    When called in parallel, this function will sum across processes
    when a controller argument is passed or the global controller is
    defined. To disable parallel summing when running in parallel, pass
    a dummy controller as follows:

    sum(array, controller=vtkmodules.vtkParallelCore.vtkDummyController()).
    """
    from .algorithms import global_func
    class SumImpl:
        def op(self):
            return algs.sum

        def mpi_op(self):
            from mpi4py import MPI
            return MPI.SUM

        def serial_composite(self, array, axis):
            res = None
            arrays = array.Arrays
            for a in arrays:
                if a is not dsa.NoneArray:
                    if res is None:
                        res = algs.sum(a, axis).astype(numpy.float64)
                    else:
                        res += algs.sum(a, axis)
            return res

        def default(self):
            return numpy.float64(0)

    return global_func(SumImpl(), array, axis, controller)

@dsa._override_numpy(numpy.max)
def max(array, axis=None, controller=None):
    """Returns the max of all values along a particular axis (dimension).
    Given an array of m tuples and n components:
    * Default is to return the max of all values in an array.
    * axis=0: Return the max values of all tuples and return a
      one tuple, n-component array.
    * axis=1: Return the max values of all components of each tuple
      and return an m-tuple, 1-component array.

    When called in parallel, this function will compute the max across
    processes when a controller argument is passed or the global controller
    is defined. To disable parallel summing when running in parallel, pass a
    dummy controller as follows:

    max(array, controller=vtkmodules.vtkParallelCore.vtkDummyController()).
    """
    from .algorithms import global_func
    class MaxImpl:
        def op(self):
            return algs.max

        def mpi_op(self):
            from mpi4py import MPI
            return MPI.MAX

        def serial_composite(self, array, axis):
            res = _apply_func2(algs.max, array, (axis,))
            clean_list = []
            for a in res:
                if a is not dsa.NoneArray:
                    clean_list.append(a)
            if clean_list is []:
                return None
            return algs.max(clean_list, axis=0).astype(numpy.float64)

        def default(self):
            return numpy.finfo(numpy.float64).min

    return global_func(MaxImpl(), array, axis, controller)

@dsa._override_numpy(numpy.min)
def min(array, axis=None, controller=None):
    """Returns the min of all values along a particular axis (dimension).
    Given an array of m tuples and n components:
    * Default is to return the min of all values in an array.
    * axis=0: Return the min values of all tuples and return a one
      tuple, n-component array.
    * axis=1: Return the min values of all components of each tuple and
      return an m-tuple, 1-component array.

    When called in parallel, this function will compute the min across processes
    when a controller argument is passed or the global controller is defined.
    To disable parallel summing when running in parallel, pass a dummy controller as follows:

    min(array, controller=vtkmodules.vtkParallelCore.vtkDummyController()).
    """
    from .algorithms import global_func
    class MinImpl:
        def op(self):
            return algs.min

        def mpi_op(self):
            from mpi4py import MPI
            return MPI.MIN

        def serial_composite(self, array, axis):
            res = _apply_func2(algs.min, array, (axis,))
            clean_list = []
            for a in res:
                if a is not dsa.NoneArray:
                    clean_list.append(a)
            if clean_list is []:
                return None
            return algs.min(clean_list, axis=0).astype(numpy.float64)

        def default(self):
            return numpy.finfo(numpy.float64).max

    return global_func(MinImpl(), array, axis, controller)

@dsa._override_numpy(numpy.all)
def all(array, axis=None, controller=None):
    """Returns True if all values of an array evaluate to True, returns
    False otherwise.
    This is useful to check if all values of an array match a certain
    condition such as:

    algorithms.all(array > 5)
    """
    from .algorithms import global_func
    class AllImpl:
        def op(self):
            return algs.all

        def mpi_op(self):
            from mpi4py import MPI
            return MPI.LAND

        def serial_composite(self, array, axis):
            res = _apply_func2(algs.all, array, (axis,))
            clean_list = []
            for a in res:
                if a is not dsa.NoneArray:
                    clean_list.append(a)
            if clean_list is []:
                return None
            return algs.all(clean_list, axis=0)

        def default(self, max_comps):
            return numpy.ones(max_comps, dtype=bool)

    return global_func(AllImpl(), array, axis, controller)

@dsa._override_numpy(numpy.mean)
def mean(array, axis=None, controller=None, size=None):
    """Returns the mean of all values along a particular axis (dimension).
    Given an array of m tuples and n components:
    * Default is to return the mean of all values in an array.
    * axis=0: Return the mean values of all components and return a one
      tuple, n-component array.
    * axis=1: Return the mean values of all components of each tuple and
      return an m-tuple, 1-component array.

    When called in parallel, this function will compute the mean across
    processes when a controller argument is passed or the global controller
    is defined. To disable parallel summing when running in parallel, pass a
    dummy controller as follows:

    mean(array, controller=vtkmodules.vtkParallelCore.vtkDummyController()).
    """
    from .algorithms import array_count
    if axis is None or axis == 0:
        if size is None:
            size = array_count(array, axis, controller)
        return sum(array, axis, controller) / size
    else:
        if type(array) == dsa.VTKCompositeDataArray:
            return apply_ufunc(algs.mean, array, (axis,))
        else:
            return algs.mean(array, axis)

@dsa._override_numpy(numpy.var)
def var(array, axis=None, controller=None):
    """Returns the variance of all values along a particular axis (dimension).
    Given an array of m tuples and n components:
    * Default is to return the variance of all values in an array.
    * axis=0: Return the variance values of all components and return a one
      tuple, n-component array.
    * axis=1: Return the variance values of all components of each tuple and
      return an m-tuple, 1-component array.

    When called in parallel, this function will compute the variance across
    processes when a controller argument is passed or the global controller
    is defined. To disable parallel summing when running in parallel, pass a
    dummy controller as follows:

    var(array, controller=vtkmodules.vtkParallelCore.vtkDummyController()).
    """
    from .algorithms import array_count
    if axis is None or axis == 0:
        size = array_count(array, axis, controller)
        tmp = array - mean(array, axis, controller, size)
        return sum(tmp*tmp, axis, controller) / size
    else:
        if type(array) == dsa.VTKCompositeDataArray:
            return apply_ufunc(algs.var, array, (axis,))
        else:
            return algs.var(array, axis)

@dsa._override_numpy(numpy.std)
def std(array, axis=None, controller=None):
    """Returns the standard deviation of all values along a particular
    axis (dimension).
    Given an array of m tuples and n components:
    * Default is to return the standard deviation of all values in an array.
    * axis=0: Return the standard deviation values of all components and
      return a one tuple, n-component array.
    * axis=1: Return the standard deviation values of all components of
      each tuple and return an m-tuple, 1-component array.

    When called in parallel, this function will compute the standard deviation
    across processes when a controller argument is passed or the global controller
    is defined. To disable parallel summing when running in parallel, pass a dummy
    controller as follows:

    std(array, controller=vtkmodules.vtkParallelCore.vtkDummyController()).
    """
    from .algorithms import sqrt
    return sqrt(var(array, axis, controller))

@dsa._override_numpy(numpy.shape)
def shape(array):
    "Returns the shape (dimensions) of an array."
    if type(array) == dsa.VTKCompositeDataArray:
        shp = None
        for a in array.Arrays:
            if a is not dsa.NoneArray:
                if shp is None:
                    shp = list(a.shape)
                else:
                    tmp = a.shape
                    if (len(shp) != len(tmp)):
                        raise ValueError("Expected arrays of same shape")
                    shp[0] += tmp[0]
                    for idx in range(1,len(tmp)):
                        if shp[idx] != tmp[idx]:
                            raise ValueError("Expected arrays of same shape")
        return tuple(shp)
    elif array is dsa.NoneArray:
        return ()
    else:
        return numpy.shape(array)

@dsa._override_numpy(numpy.flatnonzero)
def flatnonzero(array):
    """
    Return indices that are non-zero in the flattened version of the input array.
    """
    if array is dsa.NoneArray:
        return dsa.NoneArray
    elif type(array) == dsa.VTKCompositeDataArray:
        res = []
        offset = 0
        for arr in array.Arrays:
            if arr is dsa.NoneArray:
                res.append(dsa.NoneArray)
            else:
                res.append(numpy.flatnonzero(arr) + offset)
                offset += arr.size
        return dsa.VTKCompositeDataArray(res, dataset = array.DataSet)
    else:
        return numpy.flatnonzero(array)

def _like_VTKCompositeDataArray(array: dsa.VTKCompositeDataArray, kind: str,
                                dtype: Optional[numpy.dtype] = None, value: Optional[int] = None):
    """Helper to construct like-structured VTKCompositeDataArray."""
    arrays = []
    for arr in array.Arrays:
        if arr is dsa.NoneArray:
            arrays.append(dsa.NoneArray)
            continue

        new_dtype = dtype if dtype is not None else arr.dtype

        if kind == "full":
            arrays.append(numpy.full(arr.shape, value, dtype=new_dtype))
        else:
            arrays.append(numpy.empty(arr.shape, dtype=new_dtype))

    return dsa.VTKCompositeDataArray(arrays, dataset=array.DataSet, association=array.Association)

@dsa._override_numpy(numpy.zeros_like)
def zeros_like(array: dsa.VTKCompositeDataArray, dtype: Optional[numpy.dtype] = None) -> dsa.VTKCompositeDataArray:
    """Create a new VTKCompositeDataArray filled with 0 of the same shape as array."""
    return _like_VTKCompositeDataArray(array, "full", dtype, 0)

@dsa._override_numpy(numpy.ones_like)
def ones_like(array: dsa.VTKCompositeDataArray, dtype: Optional[numpy.dtype] = None) -> dsa.VTKCompositeDataArray:
    """Create a new VTKCompositeDataArray filled with 1 of the same shape as array."""
    return _like_VTKCompositeDataArray(array, "full", dtype, 1)

@dsa._override_numpy(numpy.empty_like)
def empty_like(array: dsa.VTKCompositeDataArray, dtype: Optional[numpy.dtype] = None) -> dsa.VTKCompositeDataArray:
    """Create a new uninitialized VTKCompositeDataArray of the same shape as array."""
    return _like_VTKCompositeDataArray(array, "empty", dtype)

nonzero = (dsa._override_numpy(numpy.nonzero)(_make_ufunc(numpy.nonzero)))
nonzero.__doc__ = """Return the indices of the non-zero elements of the input array.
For vtkCompositeDataArray, indices are local to each chunk."""

where = (dsa._override_numpy(numpy.where)(_make_ufunc(numpy.where)))
where.__doc__ = """Returns the location (indices) of an array where the given
expression is true. For vtkCompositeDataArray, indices are local to each chunk.
For scalars, it returns a single array of indices.
For vectors and matrices, it returns two arrays: first with tuple indices,
second with component indices. The output of this method can be used to
extract the values from the array also by using it as the index of the [] operator.

For example:

>>> algs.where(algs.array([1,2,3]) == 2)
(array([1]),)

>>> algs.where(algs.array([[1,2,3], [2,1,1]]) == 2)
(array([0, 1]), array([1, 0]))

>>> a = array([[1,2,3], [2,1,1]])
>>> indices = algs.where(a > 2)
>>> a[indices]
array([3])
"""


isin = (dsa._override_numpy(numpy.isin)(_make_ufunc(numpy.isin)))
isin.__doc__ = "Test whether each element of a 1-D array is also present in a second array."

expand_dims = dsa._override_numpy(numpy.expand_dims)(_make_dfunc(numpy.expand_dims))
expand_dims.__doc__ = """Insert a new dimension, corresponding to a given
position in the array shape. In VTK, this function's main use is to
enable an operator to work on a vector and a scalar field. For example,
say you want to divide each component of a vector by the magnitude of
that vector. You might try this:

>>> v
VTKArray([[ 1.,  1.,  1.],
       [ 1.,  1.,  1.],
       [ 1.,  1.,  1.],
       [ 1.,  1.,  1.],
       [ 1.,  1.,  1.]])
>>> algs.mag(v)
VTKArray([ 1.73205081,  1.73205081,  1.73205081,  1.73205081,  1.73205081])
>>> v / algs.mag(v)
Traceback (most recent call last):
  File "<stdin>", line 1, in <module>
ValueError: operands could not be broadcast together with shapes (5,3) (5)

The division operator does not know how to map a scalar to a vector
due to a mismatch in dimensions. This can be solved by making the
scalar a vector of 1 component (increasing its dimension to 2) as follows:

>>> v / np.expand_dims(algs.mag(v), 1)
VTKArray([[ 0.57735027,  0.57735027,  0.57735027],
       [ 0.57735027,  0.57735027,  0.57735027],
       [ 0.57735027,  0.57735027,  0.57735027],
       [ 0.57735027,  0.57735027,  0.57735027],
       [ 0.57735027,  0.57735027,  0.57735027]])"""



"""This module provides classes that allow Numpy-type access
to VTK datasets and arrays. This is best described with some examples.

To normalize a VTK array:

from vtkmodules.vtkImagingCore vtkRTAnalyticSource
import vtkmodules.numpy_interface.dataset_adapter as dsa
import vtkmodules.numpy_interface.algorithms as algs

rt = vtkRTAnalyticSource()
rt.Update()
image = dsa.WrapDataObject(rt.GetOutput())
rtdata = image.PointData['RTData']
rtmin = algs.min(rtdata)
rtmax = algs.max(rtdata)
rtnorm = (rtdata - rtmin) / (rtmax - rtmin)
image.PointData.append(rtnorm, 'RTData - normalized')
print image.GetPointData().GetArray('RTData - normalized').GetRange()

To calculate gradient:

grad= algs.gradient(rtnorm)

To access subsets:

>>> grad[0:10]
VTKArray([[ 0.10729134,  0.03763443,  0.03136338],
       [ 0.02754352,  0.03886006,  0.032589  ],
       [ 0.02248248,  0.04127144,  0.03500038],
       [ 0.02678365,  0.04357527,  0.03730421],
       [ 0.01765099,  0.04571581,  0.03944477],
       [ 0.02344007,  0.04763837,  0.04136734],
       [ 0.01089381,  0.04929155,  0.04302051],
       [ 0.01769151,  0.05062952,  0.04435848],
       [ 0.002764  ,  0.05161414,  0.04534309],
       [ 0.01010841,  0.05221677,  0.04594573]])

>>> grad[:, 0]
VTKArray([ 0.10729134,  0.02754352,  0.02248248, ..., -0.02748174,
       -0.02410045,  0.05509736])

All of this functionality is also supported for composite datasets
even though their data arrays may be spread across multiple datasets.
We have implemented a VTKCompositeDataArray class that handles many
Numpy style operators and is supported by all algorithms in the
algorithms module.

This module also provides an API to access composite datasets.
For example:

from vtkmodules.vtkCommonDataModel import vtkMultiBlockDataSet
mb = vtkMultiBlockDataSet()
mb.SetBlock(0, image.VTKObject)
mb.SetBlock(1e, image.VTKObject)
cds = dsa.WrapDataObject(mb)
for block in cds:
    print block

Note that this module implements only the wrappers for datasets
and arrays. The classes implement many useful operators. However,
to make best use of these classes, take a look at the algorithms
module.
"""
import numpy

# version 2 has many API changes from version 1
NUMPY_MAJOR_VERSION = int(numpy.__version__.split('.')[0])

from typing import Generator, Tuple, Union, List
import operator
import sys
from ..vtkCommonCore import buffer_shared
from ..util import numpy_support
from ..vtkCommonDataModel import vtkDataObject
from ..vtkCommonCore import vtkWeakReference, vtkObject
import weakref
from math import ceil

COMPOSITE_OVERRIDE = {}
def _override_numpy(numpy_function):
    """Register an __array_function__ implementation for VTKCompositeDataArray objects."""
    def decorator(func):
        COMPOSITE_OVERRIDE[numpy_function] = func
        return func
    return decorator

def reshape_append_ones (a1, a2):
    """Returns a list with the two arguments, any of them may be
    processed.  If the arguments are numpy.ndarrays, append 1s to the
    shape of the array with the smallest number of dimensions until
    the arrays have the same number of dimensions. Does nothing if the
    arguments are not ndarrays or the arrays have the same number of
    dimensions.

    """
    l = [a1, a2]
    if (isinstance(a1, numpy.ndarray) and isinstance(a2, numpy.ndarray)):
        len1 = len(a1.shape)
        len2 = len(a2.shape)
        if (len1 == len2 or len1 == 0 or len2 == 0 or
            a1.shape[0] != a2.shape[0]):
            return l
        elif (len1 < len2):
            d = len1
            maxLength = len2
            i = 0
        else:
            d = len2
            maxLength = len1
            i = 1
        while (d < maxLength):
            l[i] = numpy.expand_dims(l[i], d)
            d = d + 1
    return l

class ArrayAssociation :
    """Easy access to vtkDataObject.AttributeTypes"""
    POINT = vtkDataObject.POINT
    CELL  = vtkDataObject.CELL
    FIELD = vtkDataObject.FIELD
    ROW = vtkDataObject.ROW

class VTKObjectWrapper(object):
    """Superclass for classes that wrap VTK objects with Python objects.
    This class holds a reference to the wrapped VTK object. It also
    forwards unresolved methods to the underlying object by overloading
    __get__attr."""
    def __init__(self, vtkobject):
        self.VTKObject = vtkobject

    def __getattr__(self, name):
        "Forwards unknown attribute requests to VTK object."
        return getattr(self.VTKObject, name)

def vtkDataArrayToVTKArray(array, dataset=None):
    "Given a vtkDataArray and a dataset owning it, returns a VTKArray."
    narray = numpy_support.vtk_to_numpy(array)

    # Make arrays of 9 components into matrices. Also transpose
    # as VTK store matrices in Fortran order
    shape = narray.shape
    if len(shape) == 2 and shape[1] == 9:
        narray = narray.reshape((shape[0], 3, 3)).transpose(0, 2, 1)

    return VTKArray(narray, array=array, dataset=dataset)

def numpyTovtkDataArray(array, name="numpy_array", array_type=None):
    """Given a numpy array or a VTKArray and a name, returns a vtkDataArray.
    The resulting vtkDataArray will store a reference to the numpy array:
    the numpy array is released only when the vtkDataArray is destroyed."""
    vtkarray = numpy_support.numpy_to_vtk(array, array_type=array_type)
    vtkarray.SetName(name)
    return vtkarray

def _make_tensor_array_contiguous(array):
    if array is None:
        return None
    if array.flags.contiguous:
        return array
    array = numpy.asarray(array)
    size = array.dtype.itemsize
    strides = array.strides
    if len(strides) == 3 and strides[1]/size == 1 and strides[2]/size == 3:
        return array.transpose(0, 2, 1)
    return array

def _metaclass(mcs):
    """For compatibility between python 2 and python 3."""
    def decorator(cls):
        body = vars(cls).copy()
        body.pop('__dict__', None)
        body.pop('__weakref__', None)
        return mcs(cls.__name__, cls.__bases__, body)
    return decorator

class VTKArrayMetaClass(type):
    def __new__(mcs, name, parent, attr):
        """We overwrite numerical/comparison operators because we might need
        to reshape one of the arrays to perform the operation without
        broadcast errors. For instance:

        An array G of shape (n,3) resulted from computing the
        gradient on a scalar array S of shape (n,) cannot be added together without
        reshaping.
        G + expand_dims(S,1) works,
        G + S gives an error:
        ValueError: operands could not be broadcast together with shapes (n,3) (n,)

        This metaclass overwrites operators such that it computes this
        reshape operation automatically by appending 1s to the
        dimensions of the array with fewer dimensions.

        """
        def add_numeric_op(attr_name):
            """Create an attribute named attr_name that calls
            _numeric_op(self, other, op)."""
            def closure(self, other):
                return VTKArray._numeric_op(self, other, attr_name)
            closure.__name__ = attr_name
            attr[attr_name] = closure

        def add_default_numeric_op(op_name):
            """Adds '__[op_name]__' attribute that uses operator.[op_name]"""
            add_numeric_op("__%s__"%op_name)

        def add_reverse_numeric_op(attr_name):
            """Create an attribute named attr_name that calls
            _reverse_numeric_op(self, other, op)."""
            def closure(self, other):
                return VTKArray._reverse_numeric_op(self, other, attr_name)
            closure.__name__ = attr_name
            attr[attr_name] = closure

        def add_default_reverse_numeric_op(op_name):
            """Adds '__r[op_name]__' attribute that uses operator.[op_name]"""
            add_reverse_numeric_op("__r%s__"%op_name)

        def add_default_numeric_ops(op_name):
            """Call both add_default_numeric_op and add_default_reverse_numeric_op."""
            add_default_numeric_op(op_name)
            add_default_reverse_numeric_op(op_name)

        add_default_numeric_ops("add")
        add_default_numeric_ops("sub")
        add_default_numeric_ops("mul")
        add_default_numeric_ops("truediv")
        add_default_numeric_ops("floordiv")
        add_default_numeric_ops("mod")
        add_default_numeric_ops("pow")
        add_default_numeric_ops("lshift")
        add_default_numeric_ops("rshift")
        add_numeric_op("and")
        add_default_numeric_ops("xor")
        add_numeric_op("or")

        add_default_numeric_op("lt")
        add_default_numeric_op("le")
        add_default_numeric_op("eq")
        add_default_numeric_op("ne")
        add_default_numeric_op("ge")
        add_default_numeric_op("gt")
        return type.__new__(mcs, name, parent, attr)

@_metaclass(VTKArrayMetaClass)
class VTKArray(numpy.ndarray):
    """This is a sub-class of numpy ndarray that stores a
    reference to a vtk array as well as the owning dataset.
    The numpy array and vtk array should point to the same
    memory location."""

    def _numeric_op(self, other, attr_name):
        """Used to implement numpy-style numerical operations such as __add__,
        __mul__, etc."""
        l = reshape_append_ones(self, other)
        return getattr(numpy.ndarray, attr_name)(l[0], l[1])

    def _reverse_numeric_op(self, other, attr_name):
        """Used to implement numpy-style numerical operations such as __add__,
        __mul__, etc."""
        l = reshape_append_ones(self, other)
        return getattr(numpy.ndarray, attr_name)(l[0], l[1])

    def __new__(cls, input_array, array=None, dataset=None):
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        obj = numpy.asarray(input_array).view(cls)
        obj.Association = ArrayAssociation.FIELD
        # add the new attributes to the created instance
        obj.VTKObject = array
        if dataset:
            obj._dataset = vtkWeakReference()
            if issubclass(type(dataset), vtkObject):
                # New dataset API.
                obj._dataset.Set(dataset)
            else:
                # Old dataset type object with WrapDataObject
                obj._dataset.Set(dataset.VTKObject)
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self,obj):
        # Copy the VTK array only if the two share data
        slf = _make_tensor_array_contiguous(self)
        obj2 = _make_tensor_array_contiguous(obj)

        self.VTKObject = None
        try:
            # This line tells us that they are referring to the same buffer.
            # Much like two pointers referring to same memory location in C/C++.
            if buffer_shared(slf, obj2):
                self.VTKObject = getattr(obj, 'VTKObject', None)
        except TypeError:
            pass

        self.Association = getattr(obj, 'Association', None)
        self.DataSet = getattr(obj, 'DataSet', None)

    def __getattr__(self, name):
        "Forwards unknown attribute requests to VTK array."
        try:
            o = self.__dict__["VTKObject"]
        except KeyError:
            o = None
        if o is None:
            raise AttributeError("'%s' object has no attribute '%s'" %
                                 (self.__class__.__name__, name))
        return getattr(o, name)

    def __array_wrap__(self, out_arr, context=None, return_scalar=False):
        if return_scalar or (NUMPY_MAJOR_VERSION < 2 and out_arr.shape == ()):
            # Convert to scalar value
            return out_arr[()]
        else:
            return numpy.ndarray.__array_wrap__(self, out_arr, context)

    @property
    def DataSet(self):
        """
        Get the dataset this array is associated with. The reference to the
        dataset is held through a vtkWeakReference to ensure it doesn't prevent
        the dataset from being collected if necessary.
        """
        if hasattr(self, '_dataset') and self._dataset and self._dataset.Get():
            return WrapDataObject(self._dataset.Get())

        return  None

    @DataSet.setter
    def DataSet(self, dataset):
        """
        Set the dataset this array is associated with. The reference is held
        through a vtkWeakReference.
        """
        # Do we have dataset to store
        if dataset and dataset.VTKObject:
            # Do we need to create a vtkWeakReference
            if not hasattr(self, '_dataset') or self._dataset is None:
                self._dataset = vtkWeakReference()

            self._dataset.Set(dataset.VTKObject)
        else:
            self._dataset = None

class VTKNoneArrayMetaClass(type):
    def __new__(mcs, name, parent, attr):
        """Simplify the implementation of the numeric/logical sequence API."""
        def _add_op(attr_name, op):
            """Create an attribute named attr_name that calls
            _numeric_op(self, other, op)."""
            def closure(self, other):
                return VTKNoneArray._op(self, other, op)
            closure.__name__ = attr_name
            attr[attr_name] = closure

        def _add_default_reverse_op(op_name):
            """Adds '__r[op_name]__' attribute that uses operator.[op_name]"""
            _add_op("__r%s__"%op_name, getattr(operator, op_name))

        def _add_default_op(op_name):
            """Adds '__[op_name]__' attribute that uses operator.[op_name]"""
            _add_op("__%s__"%op_name, getattr(operator, op_name))

        def _add_default_ops(op_name):
            """Call both add_default_numeric_op and add_default_reverse_numeric_op."""
            _add_default_op(op_name)
            _add_default_reverse_op(op_name)

        _add_default_ops("add")
        _add_default_ops("sub")
        _add_default_ops("mul")
        _add_default_ops("truediv")
        _add_default_ops("floordiv")
        _add_default_ops("mod")
        _add_default_ops("pow")
        _add_default_ops("lshift")
        _add_default_ops("rshift")
        _add_op("__and__", operator.and_)
        _add_op("__rand__", operator.and_)
        _add_default_ops("xor")
        _add_op("__or__", operator.or_)
        _add_op("__ror__", operator.or_)

        _add_default_op("lt")
        _add_default_op("le")
        _add_default_op("eq")
        _add_default_op("ne")
        _add_default_op("ge")
        _add_default_op("gt")
        return type.__new__(mcs, name, parent, attr)

@_metaclass(VTKNoneArrayMetaClass)
class VTKNoneArray(object):
    """VTKNoneArray is used to represent a "void" array. An instance
    of this class (NoneArray) is returned instead of None when an
    array that doesn't exist in a DataSetAttributes is requested.
    All operations on the NoneArray return NoneArray. The main reason
    for this is to support operations in parallel where one of the
    processes may be working on an empty dataset. In such cases,
    the process is still expected to evaluate a whole expression because
    some of the functions may perform bulk MPI communication. None
    cannot be used in these instances because it cannot properly override
    operators such as __add__, __sub__ etc. This is the main raison
    d'etre for VTKNoneArray."""

    def __getitem__(self, index):
        return NoneArray

    def _op(self, other, op):
        """Used to implement numpy-style numerical operations such as __add__,
        __mul__, etc."""
        return NoneArray

    def astype(self, dtype):
        """Implements numpy array's astype method."""
        return NoneArray

    @property
    def shape(self) -> Tuple[int, ...]:
        """Return a tuple representing the shape of the none array."""
        return (0,)

NoneArray = VTKNoneArray()

class VTKCompositeDataArrayMetaClass(type):
    def __new__(mcs, name, parent, attr):
        """Simplify the implementation of the numeric/logical sequence API."""
        def add_numeric_op(attr_name, op):
            """Create an attribute named attr_name that calls
            _numeric_op(self, other, op)."""
            def closure(self, other):
                return VTKCompositeDataArray._numeric_op(self, other, op)
            closure.__name__ = attr_name
            attr[attr_name] = closure

        def add_reverse_numeric_op(attr_name, op):
            """Create an attribute named attr_name that calls
            _reverse_numeric_op(self, other, op)."""
            def closure(self, other):
                return VTKCompositeDataArray._reverse_numeric_op(self, other, op)
            closure.__name__ = attr_name
            attr[attr_name] = closure

        def add_default_reverse_numeric_op(op_name):
            """Adds '__r[op_name]__' attribute that uses operator.[op_name]"""
            add_reverse_numeric_op("__r%s__"%op_name, getattr(operator, op_name))

        def add_default_numeric_op(op_name):
            """Adds '__[op_name]__' attribute that uses operator.[op_name]"""
            add_numeric_op("__%s__"%op_name, getattr(operator, op_name))

        def add_default_numeric_ops(op_name):
            """Call both add_default_numeric_op and add_default_reverse_numeric_op."""
            add_default_numeric_op(op_name)
            add_default_reverse_numeric_op(op_name)

        add_default_numeric_ops("add")
        add_default_numeric_ops("sub")
        add_default_numeric_ops("mul")
        add_default_numeric_ops("truediv")
        add_default_numeric_ops("floordiv")
        add_default_numeric_ops("mod")
        add_default_numeric_ops("pow")
        add_default_numeric_ops("lshift")
        add_default_numeric_ops("rshift")
        add_numeric_op("__and__", operator.and_)
        add_reverse_numeric_op("__rand__", operator.and_)
        add_default_numeric_ops("xor")
        add_numeric_op("__or__", operator.or_)
        add_reverse_numeric_op("__ror__", operator.or_)

        add_default_numeric_op("lt")
        add_default_numeric_op("le")
        add_default_numeric_op("eq")
        add_default_numeric_op("ne")
        add_default_numeric_op("ge")
        add_default_numeric_op("gt")
        return type.__new__(mcs, name, parent, attr)

@_metaclass(VTKCompositeDataArrayMetaClass)
class VTKCompositeDataArray(object):
    """This class manages a set of arrays of the same name contained
    within a composite dataset. Its main purpose is to provide a
    Numpy-type interface to composite data arrays which are naturally
    nothing but a collection of vtkDataArrays. A VTKCompositeDataArray
    makes such a collection appear as a single Numpy array and support
    all array operations that this module and the associated algorithm
    module support. Note that this is not a subclass of a Numpy array
    and as such cannot be passed to native Numpy functions. Instead
    VTK modules should be used to process composite arrays.
    """

    def __init__(self, arrays = [], dataset = None, name = None,
                 association = None):
        """Construct a composite array given a container of
        arrays, a dataset, name and association. It is sufficient
        to define a container of arrays to define a composite array.
        It is also possible to initialize an array by defining
        the dataset, name and array association. In that case,
        the underlying arrays will be created lazily when they
        are needed. It is recommended to use the latter method
        when initializing from an existing composite dataset."""
        self._Arrays = arrays
        self.DataSet = dataset
        self.Name = name
        validAssociation = True
        if association == None:
            for array in self._Arrays:
                if hasattr(array, "Association"):
                    if association == None:
                        association = array.Association
                    elif array.Association and association != array.Association:
                        validAssociation = False
                        break
        if validAssociation:
            self.Association = association
        else:
            self.Association = ArrayAssociation.FIELD
        self.Initialized = False

    def __init_from_composite(self):
        if self.Initialized:
            return

        self.Initialized = True

        if self.DataSet is None or self.Name is None:
            return

        self._Arrays = []
        for ds in self.DataSet:
            self._Arrays.append(ds.GetAttributes(self.Association)[self.Name])

    def GetSize(self):
        "Returns the number of elements in the array."
        self.__init_from_composite()
        size = numpy.int64(0)
        for a in self._Arrays:
            try:
                size += a.size
            except AttributeError:
                pass
        return size

    size = property(GetSize)

    def GetArrays(self):
        """Returns the internal container of VTKArrays. If necessary,
        this will populate the array list from a composite dataset."""
        self.__init_from_composite()
        return self._Arrays

    Arrays = property(GetArrays)

    def __setitem__(self, index, value) -> None:
        """Setter overwritten to defer indexing to underlying VTKArrays.
        For the most part, this will behave like Numpy."""
        self.__init_from_composite()

        partition_sizes = [len(a) if a is not NoneArray else 0 for a in self._Arrays]
        offsets = numpy.cumsum([0] + partition_sizes)
        total_size = offsets[-1]

        if isinstance(index, VTKCompositeDataArray):
            for array, idx in zip(self._Arrays, index._Arrays):
                if array is not NoneArray:
                    array[idx] = value
        elif isinstance(index, (numpy.ndarray, list)):
            index = numpy.asarray(index)
            if index.dtype == bool and index.shape == self.shape:
                for array_id, array in enumerate(self._Arrays):
                    if array is NoneArray:
                        continue
                    start = offsets[array_id]
                    end   = offsets[array_id + 1]
                    local_mask = index[start:end]
                    array[local_mask] = value
            elif index.ndim == 0:
                self.__setitem__(int(index), value)
            elif index.ndim == 1:
                index = numpy.where(index < 0, index + total_size, index)
                for i, array in enumerate(self._Arrays):
                    if array is NoneArray:
                        continue
                    idx_mask = numpy.logical_and(index >= offsets[i], index < offsets[i+1])
                    whr = numpy.nonzero(idx_mask)
                    array[index[whr]-offsets[i]] = value
            else:
              raise IndexError(f"Unsupported index type: {type(index)}")
        else:
          if not isinstance(index, tuple):
              index = (index,)
          global_index = index[0]
          remaining_indices = index[1:]

          if isinstance(global_index, int):
              chunk_idx, local_idx = self._global_to_local_id(global_index, total_size, offsets)
              self._Arrays[chunk_idx][local_idx][tuple(remaining_indices)] = value
          elif isinstance(global_index, slice):
              for array, offset, size in zip(self._Arrays, offsets, partition_sizes):
                  if array is NoneArray:
                      continue
                  local_slice = self._slice_intersection(global_index, offset, size, total_size)
                  if local_slice is not None:
                      array[(local_slice,) + remaining_indices] = value
          else:
              raise TypeError(f"Unsupported index type: {type(index)}")

    def __getitem__(self, index) -> Union[List, "VTKCompositeDataArray", numpy.ndarray]:
        """Getter overwritten to refer indexing to underlying VTKArrays.
        For the most part, this will behave like Numpy."""
        self.__init_from_composite()

        empty = True
        for a in self._Arrays:
            if a is not NoneArray:
                empty = False
                dtype = a.dtype
                output_shape = a.shape[1:]
                break
        if len(self._Arrays) == 0 or empty:
            raise IndexError("Index out of bounds")

        partition_sizes = [len(a) if a is not NoneArray else 0 for a in self._Arrays]
        offsets = numpy.cumsum([0] + partition_sizes)
        total_size = offsets[-1]

        if isinstance(index, VTKCompositeDataArray):
            res = []
            for array, idx in zip(self._Arrays, index._Arrays):
                if array is not NoneArray:
                    res.append(array.__getitem__(idx))
                else:
                    res.append(NoneArray)
            return VTKCompositeDataArray(res, dataset=self.DataSet)

        if isinstance(index, (numpy.ndarray, list)):
            index = numpy.asarray(index)
            if index.ndim == 0:
                return self[int(index)]
            elif index.ndim == 1:
                index = numpy.where(index < 0, index + total_size, index)
                res = numpy.empty(index.shape + output_shape, dtype=dtype)
                for i, array in enumerate(self._Arrays):
                    if array is NoneArray:
                        continue
                    idx_mask = numpy.logical_and(index >= offsets[i], index < offsets[i+1])
                    whr = numpy.nonzero(idx_mask)
                    res[whr] = array[index[whr]-offsets[i]]
                return res
            else:
                raise IndexError("Only 1D Numpy or list indexing is supported for now")

        if not isinstance(index, tuple):
            index = (index,)

        global_index = index[0]
        remaining_indices = index[1:]

        if isinstance(global_index, int):
            chunk_idx, local_idx = self._global_to_local_id(global_index, total_size, offsets)
            result = self._Arrays[chunk_idx][local_idx]
            if remaining_indices:
                result = result[tuple(remaining_indices)]
            return result

        if isinstance(global_index, slice):
            step = global_index.step if global_index.step is not None else 1
            result = [NoneArray] * len(self._Arrays)
            i = 0
            for (array, offset, size) in zip(self._Arrays, offsets, partition_sizes):
                if array is NoneArray:
                    i += 1
                    continue
                local_slice = self._slice_intersection(global_index, offset, size, total_size)
                if local_slice is not None:
                    part = array[local_slice]
                    if remaining_indices:
                        # Slice all dimensions of the array except the first one
                        part = part[(slice(0, None, None),) + remaining_indices]
                    if step > 0:
                        result[i] = part
                    if step < 0:
                        result[-i-1] = part
                    i += 1
            return self.__class__(result, dataset=self.DataSet, association=self.Association)

        if isinstance(global_index, numpy.ndarray):
            return self[numpy.concatenate([array for array in index])]

        raise TypeError(f"Unsupported index type: {type(index)}")

    def _global_to_local_id(self, global_index, total_size, offsets):
        """Returns the chunk index and the local index of the chunk from the global index"""
        if global_index < 0:
            global_index += total_size
        if not (0 <= global_index < total_size):
            raise IndexError("Index out of bounds")
        i = numpy.searchsorted(offsets, global_index, side='right') - 1
        return i, global_index - offsets[i]

    def _slice_intersection(self, global_slice, chunk_offset, chunk_size, total_length):
        """Returns a local slice into the chunk for the global_slice, or None if no overlap.
        Handles both positive and negative steps."""
        start, stop, step = global_slice.indices(total_length)
        chunk_start = chunk_offset
        chunk_end   = chunk_offset + chunk_size

        if step > 0:
            # Find the first global index in the slice that is in the chunk.
            if start < chunk_start:
                offset_steps = (chunk_start - start + step - 1) // step
                first = start + offset_steps * step
            else:
                first = start
            if first >= stop or first >= chunk_end:
                return None

            # Find the last global index in the slice that is in the chunk.
            max_possible = min(stop, chunk_end)
            if first > max_possible:
                last = first
            else:
                num_steps = ceil((max_possible - first) / step)
                last = first + num_steps * step

        else:
            # Find the first global index in the slice that is in the chunk.
            if start >= chunk_end:
                first = chunk_end - 1
                remainder = (start - first) % (-step)
                if remainder != 0:
                    first = first - ( (-step) - remainder )
            else:
                first = start
            if first < chunk_start or first <= stop:
                return None

            # Find the last global index in the slice that is in the chunk.
            if chunk_start > stop:
                last = chunk_start - 1
            else:
                num_steps = ceil((start - chunk_start) / (-step))
                last = start + num_steps * step
                if last <= chunk_start:
                    last = chunk_start
            if last > first or last < stop:
                return None

        # Global to local indices
        local_first = first - chunk_offset
        local_last = last - chunk_offset
        local_last = local_last if local_last != -1 else None # For negative steps

        return slice(local_first, local_last, step)


    def _numeric_op(self, other, op):
        """Used to implement numpy-style numerical operations such as __add__,
        __mul__, etc."""
        self.__init_from_composite()
        res = []
        if type(other) == VTKCompositeDataArray:
            for a1, a2 in zip(self._Arrays, other.Arrays):
                if a1 is not NoneArray and a2 is not NoneArray:
                    l = reshape_append_ones(a1, a2)
                    res.append(op(l[0],l[1]))
                else:
                    res.append(NoneArray)
        else:
            for a in self._Arrays:
                if a is not NoneArray:
                    l = reshape_append_ones(a, other)
                    res.append(op(l[0], l[1]))
                else:
                    res.append(NoneArray)
        return VTKCompositeDataArray(
            res, dataset=self.DataSet, association=self.Association)

    def _reverse_numeric_op(self, other, op):
        """Used to implement numpy-style numerical operations such as __add__,
        __mul__, etc."""
        self.__init_from_composite()
        res = []
        if type(other) == VTKCompositeDataArray:
            for a1, a2 in zip(self._Arrays, other.Arrays):
                if a1 is not NoneArray and a2 is not NoneArray:
                    l = reshape_append_ones(a2,a1)
                    res.append(op(l[0],l[1]))
                else:
                    res.append(NoneArray)
        else:
            for a in self._Arrays:
                if a is not NoneArray:
                    l = reshape_append_ones(other, a)
                    res.append(op(l[0], l[1]))
                else:
                    res.append(NoneArray)
        return VTKCompositeDataArray(
            res, dataset=self.DataSet, association = self.Association)

    def __str__(self):
        return self.Arrays.__str__()

    def astype(self, dtype):
        """Implements numpy array's as array method."""
        res = []
        if self is not NoneArray:
            for a in self.Arrays:
                if a is NoneArray:
                    res.append(NoneArray)
                else:
                    res.append(a.astype(dtype))
        return VTKCompositeDataArray(
            res, dataset = self.DataSet, association = self.Association)

    @property
    def shape(self) -> Tuple[int, ...]:
        """Return a tuple representing the shape of the composite array."""
        if not self.Arrays:
            return (0,)
        return numpy.shape(self)

    def __len__(self) -> int:
        """Return the total number of elements in the composite array."""
        return self.shape[0]

    def __contains__(self, item) -> bool:
        """Check if the item exists in any of the non-null sub-arrays."""
        return any(item in array for array in self.Arrays if array is not NoneArray)

    def __reversed__(self) -> Generator[numpy.ndarray, None, None]:
        """Iterate over all subarrays in the composite array in backward order."""
        for array in reversed(self.Arrays):
            if array is not NoneArray:
                for subarray in reversed(array):
                    yield subarray

    def __iter__(self) -> Generator[numpy.ndarray, None, None]:
        """Iterate over all subarrays in the composite array in forward order."""
        for array in self.Arrays:
            if array is not NoneArray:
                for subarray in array:
                    yield subarray

    def __array__(self, dtype=None, copy=None) -> numpy.ndarray:
        """Convert the composite array into a single NumPy array."""
        if copy is False:
            raise RuntimeError("VTKCompositeDataArray must create a copy to be converted into a Numpy array.")
        return numpy.concatenate([numpy.asarray(a) for a in self.Arrays if a is not NoneArray], axis=0, dtype=dtype)

    def __array_function__(self, func, types, args, kwargs):
        """Implements Numpy dispatch mechanism. Functions registered in COMPOSITE_OVERRIDE
        are captured here. See:
        https://numpy.org/doc/stable/user/basics.interoperability.html#the-array-function-protocol"""

        if func not in COMPOSITE_OVERRIDE:
            return NotImplemented
        return COMPOSITE_OVERRIDE[func](*args, **kwargs)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Handles Numpy functions calls that takes a fixed number of specific inputs and outputs."""
        if method != '__call__':
            raise NotImplementedError(f"Method {method} is not supported by __array_ufunc__")

        num_chunks = len(self.Arrays)
        partition_sizes = [sub.shape[0] for sub in self.Arrays]
        total_size = sum(partition_sizes) if partition_sizes else 0
        global_indices = numpy.cumsum(partition_sizes)[:-1]

        # Split inputs arguments to each chunk
        # Basically, this returns a 2D array where the first dimension represents each chunck of the
        # composite array and the second dimension represents the data associated for each chunk
        def process_input(input_):
            if isinstance(input_, type(self)):
                if len(input_.Arrays) != num_chunks:
                    raise ValueError("Composite operands must have the same number of subarrays")
                return input_.Arrays
            elif numpy.isscalar(input_):
                return [input_] * num_chunks
            elif isinstance(input_, numpy.ndarray):
                if input_.ndim >= 1 and input_.shape[0] == total_size:
                    return list(numpy.split(input_, global_indices, axis=0))
                else:
                    return [input_] * num_chunks
            else:
                return [input_] * num_chunks

        splitted_inputs = [process_input(input_) for input_ in inputs]

        results = []
        for i in range(num_chunks):
            arg0 = splitted_inputs[0][i]
            args = [lst[i] for lst in splitted_inputs[1:]]
            if arg0 is NoneArray or (args and args[0] is NoneArray):
                results.append(NoneArray)
                continue
            result = ufunc(arg0, *args, **kwargs)
            results.append(result)

        return self.__class__(results, dataset=self.DataSet, association=self.Association)

class DataSetAttributes(VTKObjectWrapper):
    """This is a python friendly wrapper of vtkDataSetAttributes. It
    returns VTKArrays. It also provides the dictionary interface.
    Note that the stored array should have a shape that matches the number
    of elements. E.g. for a PointData, narray.shape[0] should be equal
    to dataset.GetNumberOfPoints()"""

    def __init__(self, vtkobject, dataset, association):
        super(DataSetAttributes, self).__init__(vtkobject)
        # import weakref
        # self.DataSet = weakref.ref(dataset)
        self.DataSet = dataset
        self.Association = association

    def __getitem__(self, idx):
        """Implements the [] operator. Accepts an array name or index."""
        return self.GetArray(idx)

    def GetArray(self, idx):
        "Given an index or name, returns a VTKArray."
        if isinstance(idx, int) and idx >= self.VTKObject.GetNumberOfArrays():
            raise IndexError("array index out of range")
        vtkarray = self.VTKObject.GetArray(idx)
        if not vtkarray:
            vtkarray = self.VTKObject.GetAbstractArray(idx)
            if vtkarray:
                return vtkarray
            return NoneArray
        array = vtkDataArrayToVTKArray(vtkarray, self.DataSet)
        array.Association = self.Association
        return array

    def keys(self):
        """Returns the names of the arrays as a list."""
        kys = []
        narrays = self.VTKObject.GetNumberOfArrays()
        for i in range(narrays):
            name = self.VTKObject.GetAbstractArray(i).GetName()
            if name:
                kys.append(name)
        return kys

    def values(self):
        """Returns the arrays as a list."""
        vals = []
        narrays = self.VTKObject.GetNumberOfArrays()
        for i in range(narrays):
            a = self.VTKObject.GetAbstractArray(i)
            if a.GetName():
                vals.append(a)
        return vals

    def PassData(self, other):
        "A wrapper for vtkDataSet.PassData."
        try:
            self.VTKObject.PassData(other)
        except TypeError:
            self.VTKObject.PassData(other.VTKObject)

    def append(self, narray, name):
        """Appends narray to the dataset attributes.

        If narray is a scalar, create an array with this scalar for each element.
        If narray is an array with a size not matching the array association
        (e.g. size should be equal to GetNumberOfPoints() for PointData),
        copy the input narray for each element. This is intended to ease
        initialization, typically using same 3d vector for each element.
        In any case, be careful about memory explosion."""
        if narray is NoneArray:
            # if NoneArray, nothing to do.
            return

        if self.Association == ArrayAssociation.POINT:
            arrLength = self.DataSet.GetNumberOfPoints()
        elif self.Association == ArrayAssociation.CELL:
            arrLength = self.DataSet.GetNumberOfCells()
        elif self.Association == ArrayAssociation.ROW \
          and self.DataSet.GetNumberOfColumns() > 0:
            arrLength = self.DataSet.GetNumberOfRows()
        else:
            if not isinstance(narray, numpy.ndarray):
                arrLength = 1
            else:
                arrLength = narray.shape[0]

        # if input is not a valid array (i.e. unexpected shape[0]),
        # create a new array and copy input for each element
        if not isinstance(narray, numpy.ndarray) or numpy.ndim(narray) == 0: # Scalar input
            dtype = narray.dtype if isinstance(narray, numpy.ndarray) else type(narray)
            tmparray = numpy.empty(arrLength, dtype=dtype)
            tmparray.fill(narray)
            narray = tmparray
        elif narray.shape[0] != arrLength: # Vector input
            components = 1
            for l in narray.shape:
                components *= l
            try:
                tmparray = numpy.empty((arrLength, components), dtype=narray.dtype)
            except numpy.core._exceptions._ArrayMemoryError as npErr:
                sys.stderr.write("Fail to copy input array for each dataset element: array is too big to be duplicated.\n"
                "Input should either be small enough to be duplicated for each element, or shape[0] should "
                "match number of element.\n"
                "Example of correct usage: to add a point PointData array, it is common to have\n"
                "array.shape[0] == 3 or array.shape[0] == dataset.GetNumberOfPoints()\n"
                )
                sys.stderr.write(str(type(npErr)) + "\n")
                sys.stderr.write(str(npErr))
                return

            tmparray[:] = narray.flatten()
            narray = tmparray

        shape = narray.shape

        if len(shape) == 3:
            # Array of matrices. We need to make sure the order  in memory is right.
            # If column order (c order), transpose. VTK wants row order (fortran
            # order). The deep copy later will make sure that the array is contiguous.
            # If row order but not contiguous, transpose so that the deep copy below
            # does not happen.
            size = narray.dtype.itemsize
            if (narray.strides[1]/size == 3 and narray.strides[2]/size == 1) or \
                (narray.strides[1]/size == 1 and narray.strides[2]/size == 3 and \
                 not narray.flags.contiguous):
                narray  = narray.transpose(0, 2, 1)

        # If array is not contiguous, make a deep copy that is contiguous
        if not narray.flags.contiguous:
            narray = numpy.ascontiguousarray(narray)

        # Flatten array of matrices to array of vectors
        if len(shape) == 3:
            narray = narray.reshape(shape[0], shape[1]*shape[2])

        # this handle the case when an input array is directly appended on the
        # output. We want to make sure that the array added to the output is not
        # referring to the input dataset.
        copy = VTKArray(narray)
        try:
            copy.VTKObject = narray.VTKObject
        except AttributeError: pass
        arr = numpyTovtkDataArray(copy, name)
        self.VTKObject.AddArray(arr)


class CompositeDataSetAttributes():
    """This is a python friendly wrapper for vtkDataSetAttributes for composite
    datasets. Since composite datasets themselves don't have attribute data,
    but the attribute data is associated with the leaf nodes in the composite
    dataset, this class simulates a DataSetAttributes interface by taking a
    union of DataSetAttributes associated with all leaf nodes."""

    def __init__(self, dataset, association):
        # import weakref
        # self.DataSet = weakref.ref(dataset)
        self.DataSet = dataset
        self.Association = association
        self.ArrayNames = []
        self.Arrays = {}

        # build the set of arrays available in the composite dataset. Since
        # composite datasets can have partial arrays, we need to iterate over
        # all non-null blocks in the dataset.
        self.__determine_arraynames()

    def __determine_arraynames(self):
        array_set = set()
        array_list = []
        for dataset in self.DataSet:
            dsa = dataset.GetAttributes(self.Association)
            for array_name in dsa.keys():
                if array_name not in array_set:
                    array_set.add(array_name)
                    array_list.append(array_name)
        self.ArrayNames = array_list

    def keys(self):
        """Returns the names of the arrays as a list."""
        return self.ArrayNames

    def __getitem__(self, idx):
        """Implements the [] operator. Accepts an array name."""
        return self.GetArray(idx)

    def append(self, narray, name):
        """Appends a new array to the composite dataset attributes."""
        if narray is NoneArray:
            # if NoneArray, nothing to do.
            return

        added = False
        if not isinstance(narray, VTKCompositeDataArray): # Scalar input
            for ds in self.DataSet:
                ds.GetAttributes(self.Association).append(narray, name)
                added = True
            if added:
                self.ArrayNames.append(name)
                # don't add the narray since it's a scalar. GetArray() will create a
                # VTKCompositeArray on-demand.
        else:
            for ds, array in zip(self.DataSet, narray.Arrays):
                if array is not None:
                    ds.GetAttributes(self.Association).append(array, name)
                    added = True
            if added:
                self.ArrayNames.append(name)
                self.Arrays[name] = weakref.ref(narray)

    def GetArray(self, idx):
        """Given a name, returns a VTKCompositeArray."""
        arrayname = idx
        if arrayname not in self.ArrayNames:
            return NoneArray
        if arrayname not in self.Arrays or self.Arrays[arrayname]() is None:
            array = VTKCompositeDataArray(
                dataset = self.DataSet, name = arrayname, association = self.Association)
            self.Arrays[arrayname] = weakref.ref(array)
        else:
            array = self.Arrays[arrayname]()
        return array

    def PassData(self, other):
        """Emulate PassData for composite datasets."""
        for this,that in zip(self.DataSet, other.DataSet):
            for assoc in [ArrayAssociation.POINT, ArrayAssociation.CELL, ArrayAssociation.ROW]:
                if this.HasAttributes(assoc) and that.HasAttributes(assoc):
                    this.GetAttributes(assoc).PassData(that.GetAttributes(assoc))

class CompositeDataIterator(object):
    """Wrapper for a vtkCompositeDataIterator class to satisfy
       the python iterator protocol. This iterator iterates
       over non-empty leaf nodes. To iterate over empty or
       non-leaf nodes, use the vtkCompositeDataIterator directly.
       """

    def __init__(self, cds):
        self.Iterator = cds.NewIterator()
        if self.Iterator:
            self.Iterator.UnRegister(None)
            self.Iterator.GoToFirstItem()

    def __iter__(self):
        return self

    def __next__(self):
        if not self.Iterator:
            raise StopIteration

        if self.Iterator.IsDoneWithTraversal():
            raise StopIteration
        retVal = self.Iterator.GetCurrentDataObject()
        self.Iterator.GoToNextItem()
        return WrapDataObject(retVal)

    def next(self):
        return self.__next__()

    def __getattr__(self, name):
        """Returns attributes from the vtkCompositeDataIterator."""
        return getattr(self.Iterator, name)

class MultiCompositeDataIterator(CompositeDataIterator):
    """Iterator that can be used to iterate over multiple
    composite datasets together. This iterator works only
    with arrays that were copied from an original using
    CopyStructured. The most common use case is to use
    CopyStructure, then iterate over input and output together
    while creating output datasets from corresponding input
    datasets."""
    def __init__(self, cds):
        CompositeDataIterator.__init__(self, cds[0])
        self.Datasets = cds

    def __next__(self):
        if not self.Iterator:
            raise StopIteration

        if self.Iterator.IsDoneWithTraversal():
            raise StopIteration
        retVal = []
        retVal.append(WrapDataObject(self.Iterator.GetCurrentDataObject()))
        if len(self.Datasets) > 1:
            for cd in self.Datasets[1:]:
                retVal.append(WrapDataObject(cd.GetDataSet(self.Iterator)))
        self.Iterator.GoToNextItem()
        return retVal

    def next(self):
        return self.__next__()

class DataObject(VTKObjectWrapper):
    """A wrapper for vtkDataObject that makes it easier to access FielData
    arrays as VTKArrays
    """

    def GetAttributes(self, type):
        """Returns the attributes specified by the type as a DataSetAttributes
         instance."""
        if type == ArrayAssociation.FIELD:
            return self.GetFieldData()
        return DataSetAttributes(self.VTKObject.GetAttributesAsFieldData(type), self, type)

    def HasAttributes(self, type):
        "Returns if current object support this attributes type"
        return type == ArrayAssociation.FIELD

    def GetFieldData(self):
        "Returns the field data as a DataSetAttributes instance."
        return DataSetAttributes(self.VTKObject.GetFieldData(), self, ArrayAssociation.FIELD)

    FieldData = property(GetFieldData, None, None, "This property returns the field data of a data object.")

class Table(DataObject):
    """A wrapper for vtkTable that makes it easier to access RowData array as
    VTKArrays
    """
    def GetRowData(self):
        "Returns the row data as a DataSetAttributes instance."
        return self.GetAttributes(ArrayAssociation.ROW)

    def HasAttributes(self, type):
        "Returns if current object support this attributes type"
        return type == ArrayAssociation.ROW or DataObject.HasAttributes(self, type)

    RowData = property(GetRowData, None, None, "This property returns the row data of the table.")

class HyperTreeGrid(DataObject):
    """A wrapper for vtkHyperTreeGrid that makes it easier to access CellData
    arrays as VTKArrays.
    """
    def GetCellData(self):
        "Returns the cell data as DataSetAttributes instance."
        return self.GetAttributes(ArrayAssociation.CELL)

    def HasAttributes(self, type):
        "Returns if current object support this attributes type"
        return type == ArrayAssociation.CELL or DataObject.HasAttributes(self, type)

    CellData = property(GetCellData, None, None, "This property returns the cell data of the hypertree grid.")

class CompositeDataSet(DataObject):
    """A wrapper for vtkCompositeData and subclasses that makes it easier
    to access Point/Cell/Field data as VTKCompositeDataArrays. It also
    provides a Python type iterator."""

    def __init__(self, vtkobject):
        DataObject.__init__(self, vtkobject)
        self._PointData = None
        self._CellData = None
        self._FieldData = None
        self._GlobalData = None
        self._Points = None

    def __iter__(self):
        "Creates an iterator for the contained datasets."
        return CompositeDataIterator(self)

    def GetNumberOfElements(self, assoc):
        """Returns the total number of cells or points depending
        on the value of assoc which can be ArrayAssociation.POINT or
        ArrayAssociation.CELL."""
        result = 0
        for dataset in self:
            result += dataset.GetNumberOfElements(assoc)
        return int(result)

    def GetNumberOfPoints(self):
        """Returns the total number of points of all datasets
        in the composite dataset. Note that this traverses the
        whole composite dataset every time and should not be
        called repeatedly for large composite datasets."""
        return self.GetNumberOfElements(ArrayAssociation.POINT)

    def GetNumberOfCells(self):
        """Returns the total number of cells of all datasets
        in the composite dataset. Note that this traverses the
        whole composite dataset every time and should not be
        called repeatedly for large composite datasets."""
        return self.GetNumberOfElements(ArrayAssociation.CELL)

    def GetAttributes(self, type):
        """Returns the attributes specified by the type as a
        CompositeDataSetAttributes instance."""
        return CompositeDataSetAttributes(self, type)

    def HasAttributes(self, type):
        "Returns true if every leaves of current composite object support this attributes type"
        for dataset in self:
            if not dataset.HasAttributes(type):
                return False

        return True

    def GetPointData(self):
        "Returns the point data as a CompositeDataSetAttributes instance."
        if self._PointData is None or self._PointData() is None:
            pdata = self.GetAttributes(ArrayAssociation.POINT)
            self._PointData = weakref.ref(pdata)
        return self._PointData()

    def GetCellData(self):
        "Returns the cell data as a CompositeDataSetAttributes instance."
        if self._CellData is None or self._CellData() is None:
            cdata = self.GetAttributes(ArrayAssociation.CELL)
            self._CellData = weakref.ref(cdata)
        return self._CellData()

    def GetFieldData(self):
        """
        "Returns the field data as a CompositeDataSetAttributes instance."
        """
        if self._FieldData is None or self._FieldData() is None:
            fdata = self.GetAttributes(ArrayAssociation.FIELD)
            self._FieldData = weakref.ref(fdata)
        return self._FieldData()

    def GetGlobalData(self):
        "Returns the global data (field data of the root) as a DataSetAttributes instance."
        if self._GlobalData is None or self._GlobalData() is None:
            gdata = super(CompositeDataSet, self).GetFieldData()
            self._GlobalData = weakref.ref(gdata)
        return self._GlobalData()

    def GetPoints(self):
        "Returns the points as a VTKCompositeDataArray instance."
        if self._Points is None or self._Points() is None:
            pts = []
            for ds in self:
                try:
                    _pts = ds.Points
                except AttributeError:
                    _pts = None

                if _pts is None:
                    pts.append(NoneArray)
                else:
                    pts.append(_pts)
            if len(pts) == 0 or all([a is NoneArray for a in pts]):
                cpts = NoneArray
            else:
                cpts = VTKCompositeDataArray(pts, dataset=self)
            self._Points = weakref.ref(cpts)
        return self._Points()

    PointData = property(GetPointData, None, None, "This property returns the point data of the leafs of a composite dataset.")
    CellData = property(GetCellData, None, None, "This property returns the cell data of the leafs of a composite dataset.")
    FieldData = property(GetFieldData, None, None, "This property returns the field data of the leafs of a composite dataset.")
    GlobalData = property(GetGlobalData, None, None, "This property returns the global data, i.e. field data of the root of a composite dataset.")
    Points = property(GetPoints, None, None, "This property returns the points of the dataset.")

class DataSet(DataObject):
    """This is a python friendly wrapper of a vtkDataSet that defines
    a few useful properties."""

    def GetPointData(self):
        "Returns the point data as a DataSetAttributes instance."
        return self.GetAttributes(ArrayAssociation.POINT)

    def GetCellData(self):
        "Returns the cell data as a DataSetAttributes instance."
        return self.GetAttributes(ArrayAssociation.CELL)

    def HasAttributes(self, type):
        "Returns if current object support this attributes type"
        return type == ArrayAssociation.POINT or type == ArrayAssociation.CELL or DataObject.HasAttributes(self, type)

    PointData = property(GetPointData, None, None, "This property returns the point data of the dataset.")
    CellData = property(GetCellData, None, None, "This property returns the cell data of a dataset.")

class PointSet(DataSet):
    """This is a python friendly wrapper of a vtkPointSet that defines
    a few useful properties."""
    def GetPoints(self):
        """Returns the points as a VTKArray instance. Returns None if the
        dataset has implicit points."""
        if not self.VTKObject.GetPoints():
            return None
        array = vtkDataArrayToVTKArray(
            self.VTKObject.GetPoints().GetData(), self)
        array.Association = ArrayAssociation.POINT
        return array

    def SetPoints(self, pts):
        """Given a VTKArray instance, sets the points of the dataset."""
        from ..vtkCommonCore import vtkPoints
        if isinstance(pts, vtkPoints):
            p = pts
        else:
            pts = numpyTovtkDataArray(pts)
            p = vtkPoints()
            p.SetData(pts)
        self.VTKObject.SetPoints(p)

    Points = property(GetPoints, SetPoints, None, "This property returns the point coordinates of dataset.")

class PolyData(PointSet):
    """This is a python friendly wrapper of a vtkPolyData that defines
    a few useful properties."""

    def GetPolygons(self):
        """Returns the polys as a VTKArray instance."""
        if not self.VTKObject.GetPolys():
            return None
        return vtkDataArrayToVTKArray(
            self.VTKObject.GetPolys().GetData(), self)

    Polygons = property(GetPolygons, None, None, "This property returns the connectivity of polygons.")

class UnstructuredGrid(PointSet):
    """This is a python friendly wrapper of a vtkUnstructuredGrid that defines
    a few useful properties."""

    def GetCellTypes(self):
        """Returns the cell types as a VTKArray instance."""
        if not self.VTKObject.GetCellTypes():
            return None
        return vtkDataArrayToVTKArray(
            self.VTKObject.GetCellTypes(), self)

    def GetCellLocations(self):
        """Returns the cell locations as a VTKArray instance."""
        if not self.VTKObject.GetCellLocationsArray():
            return None
        return vtkDataArrayToVTKArray(
            self.VTKObject.GetCellLocationsArray(), self)

    def GetCells(self):
        """Returns the cells as a VTKArray instance."""
        if not self.VTKObject.GetCells():
            return None
        return vtkDataArrayToVTKArray(
            self.VTKObject.GetCells().GetData(), self)

    def SetCells(self, cellTypes, cellLocations, cells):
        """Given cellTypes, cellLocations, cells as VTKArrays,
        populates the unstructured grid data structures."""
        from ..util.vtkConstants import VTK_ID_TYPE
        from ..vtkCommonDataModel import vtkCellArray
        cellTypes = numpyTovtkDataArray(cellTypes)
        cellLocations = numpyTovtkDataArray(cellLocations, array_type=VTK_ID_TYPE)
        cells = numpyTovtkDataArray(cells, array_type=VTK_ID_TYPE)
        ca = vtkCellArray()
        ca.SetCells(cellTypes.GetNumberOfTuples(), cells)
        self.VTKObject.SetCells(cellTypes, cellLocations, ca)

    CellTypes = property(GetCellTypes, None, None, "This property returns the types of cells.")
    CellLocations = property(GetCellLocations, None, None, "This property returns the locations of cells.")
    Cells = property(GetCells, None, None, "This property returns the connectivity of cells.")

class Graph(DataObject):
    """This is a python friendly wrapper of a vtkGraph that defines
    a few useful properties."""

    def GetVertexData(self):
        "Returns the vertex data as a DataSetAttributes instance."
        return self.GetAttributes(ArrayAssociation.VERTEX)

    def GetEdgeData(self):
        "Returns the edge data as a DataSetAttributes instance."
        return self.GetAttributes(ArrayAssociation.EDGE)

    VertexData = property(GetVertexData, None, None, "This property returns the vertex data of the graph.")
    EdgeData = property(GetEdgeData, None, None, "This property returns the edge data of the graph.")

class Molecule(DataObject):
    """This is a python friendly wrapper of a vtkMolecule that defines
    a few useful properties."""
    def GetAtomData(self):
        "Returns the atom data as a DataSetAttributes instance."
        return self.GetVertexData()

    def GetBondData(self):
        "Returns the bond data as a DataSetAttributes instance."
        return self.GetEdgeData()

    AtomData = property(GetAtomData, None, None, "This property returns the atom data of the molecule.")
    BondData = property(GetBondData, None, None, "This property returns the bond data of the molecule.")

def WrapDataObject(ds):
    """Returns a Numpy friendly wrapper of a vtkDataObject."""
    if ds.IsA("vtkPolyData"):
        return PolyData(ds)
    elif ds.IsA("vtkUnstructuredGrid"):
        return UnstructuredGrid(ds)
    elif ds.IsA("vtkPointSet"):
        return PointSet(ds)
    elif ds.IsA("vtkDataSet"):
        return DataSet(ds)
    elif ds.IsA("vtkCompositeDataSet"):
        return CompositeDataSet(ds)
    elif ds.IsA("vtkTable"):
        return Table(ds)
    elif ds.IsA("vtkMolecule"):
        return Molecule(ds)
    elif ds.IsA("vtkGraph"):
        return Table(ds)
    elif ds.IsA("vtkHyperTreeGrid"):
        return HyperTreeGrid(ds)
    elif ds.IsA("vtkDataObject"):
        return DataObject(ds)
